import pytest

import mutf8.mutf8 as pymutf8
import mutf8.cmutf8 as cmutf8


@pytest.mark.parametrize('group,decoder', (
    ('pymutf8', pymutf8.decode_modified_utf8),
    ('cmutf8', cmutf8.decode_modified_utf8)
))
@pytest.mark.slow
def test_decode(group, decoder, benchmark):
    """Compare the performance of the python and C decoders."""
    benchmark.group = 'MUTF-8 Decoding'
    benchmark.extra_info['group'] = group
    benchmark(decoder, b'\xED\xA0\xBD\xED\xB8\x88')


@pytest.mark.parametrize('group,encoder', (
    ('pymutf8', pymutf8.encode_modified_utf8),
    ('cmutf8', cmutf8.encode_modified_utf8)
))
@pytest.mark.slow
def test_encode(group, encoder, benchmark):
    """Compare the performance of the python and C encoders."""
    benchmark.group = 'MUTF-8 Encoding'
    benchmark.extra_info['group'] = group
    benchmark(encoder, '\U0001F608')



"""
Utility methods for handling oddities in character encoding encountered
when parsing and writing JVM ClassFiles or object serialization archives.

MUTF-8 is the same as CESU-8, but with different encoding for 0x00 bytes.

.. note::

    http://bugs.python.org/issue2857 was an attempt in 2008 to get support
    for MUTF-8/CESU-8 into the python core.
"""


try:
    from mutf8.cmutf8 import decode_modified_utf8, encode_modified_utf8
except ImportError:
    from mutf8.mutf8 import decode_modified_utf8, encode_modified_utf8


# Shut up linters.
ALL_IMPORTS = [decode_modified_utf8, encode_modified_utf8]



import pytest

import mutf8.mutf8 as pymutf8
import mutf8.cmutf8 as cmutf8


def pytest_addoption(parser):
    parser.addoption(
        '--runslow',
        action='store_true',
        default=False,
        help='run slow tests'
    )


def pytest_configure(config):
    config.addinivalue_line('markers', 'slow: mark test as slow to run')


def pytest_collection_modifyitems(config, items):
    if config.getoption('--runslow'):
        # --runslow given in cli: do not skip slow tests
        return

    skip_slow = pytest.mark.skip(reason='need --runslow option to run')
    for item in items:
        if 'slow' in item.keywords:
            item.add_marker(skip_slow)


@pytest.fixture(params=[pymutf8, cmutf8])
def module(request):
    yield request.param


@pytest.fixture()
def decoder(module):
    yield module.decode_modified_utf8


@pytest.fixture()
def encoder(module):
    yield module.encode_modified_utf8



def test_issue_1(encoder, decoder):
    """
    Ensure we do not regress on https://github.com/TkTech/mutf8/issues/1.

    Two issues found here:

        1. Python-based decoder could hit an infinite loop, since it didn't
           *always* increment s_ix on an iteration.
        2. C & Python decoders would incorrectly decode the `sample` below,
           because the logic for surrogate pair decoding made the incorrect
           assumption that we could short-circuit when b1 == 0xED.
    """

    # b'[\xea\xb0\x80 \xeb\x82\x98 \xeb\x8b\xa4 \xeb\x9d\xbc \xeb\xa7\x88
    # \xeb\xb0\x94  \xec\x82\xac  \xec\x95\x84\xec\x9e\x90  \xec\xb0\xa8
    # \xec\xb9\xb4 \xed\x83\x80 \xed\x8c\x8c \xed\x95\x98]'
    sample = (
        u'[\uAC00 \uB098 \uB2E4 \uB77C \uB9C8 \uBC14  \uC0AC  \uC544'
        u'\uC790  \uCC28 \uCE74 \uD0C0 \uD30C \uD558]'
    )

    encoded = encoder(sample)
    decoded = decoder(encoded)

    assert sample == decoded


def test_issue_3(encoder, decoder):
    """
    Underallocation due to an incorrect assumption on the maximum expansion
    of an encoded string.
    """
    str = '黑人抬棺組裝包'
    assert decoder(encoder(str)) == str



def decode_modified_utf8(s: bytes) -> str:
    """
    Decodes a bytestring containing modified UTF-8 as defined in section
    4.4.7 of the JVM specification.

    :param s: bytestring to be converted.
    :returns: A unicode representation of the original string.
    """
    s_out = []
    s_len = len(s)
    s_ix = 0

    while s_ix < s_len:
        b1 = s[s_ix]
        s_ix += 1

        if b1 == 0:
            raise UnicodeDecodeError(
                'mutf-8',
                s,
                s_ix - 1,
                s_ix,
                'Embedded NULL byte in input.'
            )
        if b1 < 0x80:
            # ASCII/one-byte codepoint.
            s_out.append(chr(b1))
        elif (b1 & 0xE0) == 0xC0:
            # Two-byte codepoint.
            if s_ix >= s_len:
                raise UnicodeDecodeError(
                        'mutf-8',
                        s,
                        s_ix - 1,
                        s_ix,
                        '2-byte codepoint started, but input too short to'
                        ' finish.'
                    )

            s_out.append(
                chr(
                    (b1 & 0x1F) << 0x06 |
                    (s[s_ix] & 0x3F)
                )
            )
            s_ix += 1
        elif (b1 & 0xF0) == 0xE0:
            # Three-byte codepoint.
            if s_ix + 1 >= s_len:
                raise UnicodeDecodeError(
                        'mutf-8',
                        s,
                        s_ix - 1,
                        s_ix,
                        '3-byte or 6-byte codepoint started, but input too'
                        ' short to finish.'
                    )

            b2 = s[s_ix]
            b3 = s[s_ix + 1]

            if b1 == 0xED and (b2 & 0xF0) == 0xA0:
                # Possible six-byte codepoint.
                if s_ix + 4 >= s_len:
                    raise UnicodeDecodeError(
                            'mutf-8',
                            s,
                            s_ix - 1,
                            s_ix,
                            '3-byte or 6-byte codepoint started, but input too'
                            ' short to finish.'
                        )

                b4 = s[s_ix + 2]
                b5 = s[s_ix + 3]
                b6 = s[s_ix + 4]

                if b4 == 0xED and (b5 & 0xF0) == 0xB0:
                    # Definite six-byte codepoint.
                    s_out.append(
                        chr(
                            0x10000 |
                            (b2 & 0x0F) << 0x10 |
                            (b3 & 0x3F) << 0x0A |
                            (b5 & 0x0F) << 0x06 |
                            (b6 & 0x3F)
                        )
                    )
                    s_ix += 5
                    continue

            s_out.append(
                chr(
                    (b1 & 0x0F) << 0x0C |
                    (b2 & 0x3F) << 0x06 |
                    (b3 & 0x3F)
                )
            )
            s_ix += 2
        else:
            raise RuntimeError

    return u''.join(s_out)


def encode_modified_utf8(u: str) -> bytes:
    """
    Encodes a unicode string as modified UTF-8 as defined in section 4.4.7
    of the JVM specification.

    :param u: unicode string to be converted.
    :returns: A decoded bytearray.
    """
    final_string = bytearray()

    for c in (ord(char) for char in u):
        if c == 0x00:
            # NULL byte encoding shortcircuit.
            final_string.extend([0xC0, 0x80])
        elif c <= 0x7F:
            # ASCII
            final_string.append(c)
        elif c <= 0x7FF:
            # Two-byte codepoint.
            final_string.extend([
                (0xC0 | (0x1F & (c >> 0x06))),
                (0x80 | (0x3F & c))
            ])
        elif c <= 0xFFFF:
            # Three-byte codepoint.
            final_string.extend([
                (0xE0 | (0x0F & (c >> 0x0C))),
                (0x80 | (0x3F & (c >> 0x06))),
                (0x80 | (0x3F & c))
            ])
        else:
            # Six-byte codepoint.
            final_string.extend([
                0xED,
                0xA0 | ((c >> 0x10) & 0x0F),
                0x80 | ((c >> 0x0A) & 0x3f),
                0xED,
                0xb0 | ((c >> 0x06) & 0x0f),
                0x80 | (c & 0x3f)
            ])

    return bytes(final_string)



import os
import os.path

from setuptools import setup, find_packages, Extension


root = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(root, 'README.md'), 'rb') as readme:
    long_description = readme.read().decode('utf-8')


setup(
    name='mutf8',
    packages=find_packages(),
    version='1.0.6',
    description='Fast MUTF-8 encoder & decoder',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Tyler Kennedy',
    author_email='tk@tkte.ch',
    url='http://github.com/TkTech/mutf8',
    keywords=['mutf-8', 'cesu-8', 'jvm'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
    ],
    extras_require={
        'test': [
            'pytest',
            'pytest-benchmark'
        ]
    },
    ext_modules=[
        Extension(
            'mutf8.cmutf8',
            ['mutf8/cmutf8.c'],
            language='c',
            optional=True
        )
    ]
)



import pytest


def _pairs(*args):
    for a, b in args:
        yield chr(a), bytearray(b)


def test_decode_bad_mutf8(decoder):
    """Ensure we do the right thing when we encounter invalid MUTF-8."""
    # There should never be a null byte in a MUTF-8 string. It's the
    # entire point of using MUTF-8.
    with pytest.raises(UnicodeDecodeError) as excinfo:
        decoder(b'\x00')

    assert excinfo.value.encoding == 'mutf-8'
    assert 'Embedded NULL' in excinfo.value.reason

    # Start of a two-byte codepoint without the sibling.
    with pytest.raises(UnicodeDecodeError) as excinfo:
        decoder(b'\xC2')

    assert excinfo.value.encoding == 'mutf-8'
    assert '2-byte' in excinfo.value.reason

    # Start of a six-byte codepoint without the sibling.
    with pytest.raises(UnicodeDecodeError) as excinfo:
        decoder(b'\xED')

    assert excinfo.value.encoding == 'mutf-8'
    assert '6-byte' in excinfo.value.reason

    # Start of a three-byte codepoint without the sibling.
    with pytest.raises(UnicodeDecodeError) as excinfo:
        decoder(b'\xE2')

    assert excinfo.value.encoding == 'mutf-8'
    assert '3-byte' in excinfo.value.reason


def test_two_byte(module):
    """
    Test two-byte encoding and decoding.

    Test data taken from py2jdbc.
    """
    pairs = _pairs(
        (0x0080, (0xc2, 0x80)),
        (0x0081, (0xc2, 0x81)),
        (0x0082, (0xc2, 0x82)),
        (0x0084, (0xc2, 0x84)),
        (0x0088, (0xc2, 0x88)),
        (0x0090, (0xc2, 0x90)),
        (0x00a0, (0xc2, 0xa0)),
        (0x00c0, (0xc3, 0x80)),
        (0x0180, (0xc6, 0x80)),
        (0x0280, (0xca, 0x80)),
        (0x0480, (0xd2, 0x80)),
        (0x0481, (0xd2, 0x81)),
        (0x0483, (0xd2, 0x83)),
        (0x0487, (0xd2, 0x87)),
        (0x048f, (0xd2, 0x8f)),
        (0x049f, (0xd2, 0x9f)),
        (0x04af, (0xd2, 0xaf)),
        (0x04bf, (0xd2, 0xbf)),
        (0x04ff, (0xd3, 0xbf)),
        (0x05ff, (0xd7, 0xbf)),
        (0x05ff, (0xd7, 0xbf)),
        (0x07ff, (0xdf, 0xbf))
    )

    for decoded, original in pairs:
        assert module.decode_modified_utf8(original) == decoded
        assert module.encode_modified_utf8(decoded) == original


def test_three_byte(module):
    """
    Test three-byte encoding and decoding.

    Test data taken from py2jdbc.
    """
    pairs = _pairs(
        (0x0800, (0xe0, 0xa0, 0x80)),
        (0x0801, (0xe0, 0xa0, 0x81)),
        (0x0802, (0xe0, 0xa0, 0x82)),
        (0x0804, (0xe0, 0xa0, 0x84)),
        (0x0808, (0xe0, 0xa0, 0x88)),
        (0x0810, (0xe0, 0xa0, 0x90)),
        (0x0820, (0xe0, 0xa0, 0xa0)),
        (0x0840, (0xe0, 0xa1, 0x80)),
        (0x0880, (0xe0, 0xa2, 0x80)),
        (0x0900, (0xe0, 0xa4, 0x80)),
        (0x0a00, (0xe0, 0xa8, 0x80)),
        (0x0c00, (0xe0, 0xb0, 0x80)),
        (0x1800, (0xe1, 0xa0, 0x80)),
        (0x2800, (0xe2, 0xa0, 0x80)),
        (0x4800, (0xe4, 0xa0, 0x80)),
        (0x8800, (0xe8, 0xa0, 0x80)),
        (0x8801, (0xe8, 0xa0, 0x81)),
        (0x8803, (0xe8, 0xa0, 0x83)),
        (0x8807, (0xe8, 0xa0, 0x87)),
        (0x880f, (0xe8, 0xa0, 0x8f)),
        (0x881f, (0xe8, 0xa0, 0x9f)),
        (0x883f, (0xe8, 0xa0, 0xbf)),
        (0x887f, (0xe8, 0xa1, 0xbf)),
        (0x88ff, (0xe8, 0xa3, 0xbf)),
        (0x89ff, (0xe8, 0xa7, 0xbf)),
        (0x8bff, (0xe8, 0xaf, 0xbf)),
        (0x8fff, (0xe8, 0xbf, 0xbf)),
        (0x9fff, (0xe9, 0xbf, 0xbf)),
        (0xbfff, (0xeb, 0xbf, 0xbf)),
        (0xffff, (0xef, 0xbf, 0xbf))
    )

    for decoded, original in pairs:
        assert module.decode_modified_utf8(original) == decoded
        assert module.encode_modified_utf8(decoded) == original


def test_six_byte(module):
    """
    Test six-byte encoding and decoding.

    Test data taken from py2jdbc.
    """
    pairs = _pairs(
        (0x10000, (0xed, 0xa1, 0x80, 0xed, 0xb0, 0x80)),
        (0x10001, (0xed, 0xa1, 0x80, 0xed, 0xb0, 0x81)),
        (0x10002, (0xed, 0xa1, 0x80, 0xed, 0xb0, 0x82)),
        (0x10004, (0xed, 0xa1, 0x80, 0xed, 0xb0, 0x84)),
        (0x10008, (0xed, 0xa1, 0x80, 0xed, 0xb0, 0x88)),
        (0x10010, (0xed, 0xa1, 0x80, 0xed, 0xb0, 0x90)),
        (0x10020, (0xed, 0xa1, 0x80, 0xed, 0xb0, 0xa0)),
        (0x10040, (0xed, 0xa1, 0x80, 0xed, 0xb1, 0x80)),
        (0x10080, (0xed, 0xa1, 0x80, 0xed, 0xb2, 0x80)),
        (0x10100, (0xed, 0xa1, 0x80, 0xed, 0xb4, 0x80)),
        (0x10200, (0xed, 0xa1, 0x80, 0xed, 0xb8, 0x80)),
        (0x10400, (0xed, 0xa1, 0x81, 0xed, 0xb0, 0x80)),
        (0x10800, (0xed, 0xa1, 0x82, 0xed, 0xb0, 0x80)),
        (0x11000, (0xed, 0xa1, 0x84, 0xed, 0xb0, 0x80)),
        (0x12000, (0xed, 0xa1, 0x88, 0xed, 0xb0, 0x80)),
        (0x14000, (0xed, 0xa1, 0x90, 0xed, 0xb0, 0x80)),
        (0x18000, (0xed, 0xa1, 0xa0, 0xed, 0xb0, 0x80)),
        (0x30000, (0xed, 0xa3, 0x80, 0xed, 0xb0, 0x80)),
        (0x50000, (0xed, 0xa5, 0x80, 0xed, 0xb0, 0x80)),
        (0x90000, (0xed, 0xa9, 0x80, 0xed, 0xb0, 0x80)),
        (0x10003, (0xed, 0xa1, 0x80, 0xed, 0xb0, 0x83)),
        (0x10007, (0xed, 0xa1, 0x80, 0xed, 0xb0, 0x87)),
        (0x1000f, (0xed, 0xa1, 0x80, 0xed, 0xb0, 0x8f)),
        (0x1001f, (0xed, 0xa1, 0x80, 0xed, 0xb0, 0x9f)),
        (0x1003f, (0xed, 0xa1, 0x80, 0xed, 0xb0, 0xbf)),
        (0x1007f, (0xed, 0xa1, 0x80, 0xed, 0xb1, 0xbf)),
        (0x100ff, (0xed, 0xa1, 0x80, 0xed, 0xb3, 0xbf)),
        (0x101ff, (0xed, 0xa1, 0x80, 0xed, 0xb7, 0xbf)),
        (0x103ff, (0xed, 0xa1, 0x80, 0xed, 0xbf, 0xbf)),
        (0x107ff, (0xed, 0xa1, 0x81, 0xed, 0xbf, 0xbf)),
        (0x10fff, (0xed, 0xa1, 0x83, 0xed, 0xbf, 0xbf)),
        (0x11fff, (0xed, 0xa1, 0x87, 0xed, 0xbf, 0xbf)),
        (0x13fff, (0xed, 0xa1, 0x8f, 0xed, 0xbf, 0xbf)),
        (0x17fff, (0xed, 0xa1, 0x9f, 0xed, 0xbf, 0xbf)),
        (0x1ffff, (0xed, 0xa1, 0xbf, 0xed, 0xbf, 0xbf)),
        (0x3ffff, (0xed, 0xa3, 0xbf, 0xed, 0xbf, 0xbf)),
        (0x7ffff, (0xed, 0xa7, 0xbf, 0xed, 0xbf, 0xbf)),
        (0xfffff, (0xed, 0xaf, 0xbf, 0xed, 0xbf, 0xbf))
    )

    for decoded, original in pairs:
        assert module.decode_modified_utf8(original) == decoded
        assert module.encode_modified_utf8(decoded) == original
